@@ -6,7 +6,6 @@ import textwrap
 import asyncio
 import fsspec
 
-from base64 import b64encode, b64decode
 import google.auth as gauth
 import google.auth.compute_engine
 import google.auth.credentials
@@ -15,7 +14,6 @@ from google.oauth2.credentials import Credentials
 from google_auth_oauthlib.flow import InstalledAppFlow
 from google.oauth2 import service_account
 from google.auth.transport.requests import Request
-from hashlib import md5
 import io
 import json
 import logging
@@ -32,8 +30,10 @@ import weakref
 
 from requests.exceptions import RequestException, ProxyError
 from fsspec.asyn import sync_wrapper, sync, AsyncFileSystem
+from fsspec.utils import stringify_path
 from fsspec.implementations.http import get_client
 from .utils import ChecksumError, HttpError, is_retriable
+from .checkers import get_consistency_checker, MD5Checker
 from . import __version__ as version
 
 logger = logging.getLogger("gcsfs")
@@ -236,6 +236,7 @@ class GCSFileSystem(AsyncFileSystem):
         requester_pays=False,
         asynchronous=False,
         loop=None,
+        callback_timeout=None,
         **kwargs,
     ):
         super().__init__(
@@ -263,13 +264,30 @@ class GCSFileSystem(AsyncFileSystem):
         self.cache_timeout = cache_timeout or kwargs.pop("listings_expiry_time", None)
         self.requests_timeout = requests_timeout
         self.check_credentials = check_connection
+        self.callback_timeout = callback_timeout
         if not asynchronous:
-            self._session = sync(self.loop, get_client)
+            self._session = sync(
+                self.loop, get_client, callback_timeout=self.callback_timeout
+            )
             weakref.finalize(self, sync, self.loop, self.session.close)
         else:
             self._session = None
         self.connect(method=token)
 
+    @classmethod
+    def _strip_protocol(cls, path):
+        if isinstance(path, list):
+            return [cls._strip_protocol(p) for p in path]
+        path = stringify_path(path)
+        protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
+        for protocol in protos:
+            if path.startswith(protocol + "://"):
+                path = path[len(protocol) + 3 :]
+            elif path.startswith(protocol + "::"):
+                path = path[len(protocol) + 2 :]
+        # use of root_marker to make minimum required path, e.g., "/"
+        return path or cls.root_marker
+
     @property
     def session(self):
         if self._session is None:
@@ -369,6 +387,8 @@ class GCSFileSystem(AsyncFileSystem):
         else:
             raise ValueError("Token format not understood")
         self.credentials = credentials
+        if self.credentials.valid:
+            self.credentials.apply(self.heads)
 
     def maybe_refresh(self):
         # this uses requests and is blocking
@@ -429,9 +449,11 @@ class GCSFileSystem(AsyncFileSystem):
                         self.ls("anaconda-public-data")
                     logger.debug("Connected with method %s", meth)
                     break
-                except:  # noqa: E722
+                except Exception as e:  # noqa: E722
                     # TODO: catch specific exceptions
-                    logger.debug('Connection with method "%s" failed' % meth)
+                    logger.debug(
+                        'Connection with method "%s" failed' % meth, exc_info=e
+                    )
         else:
             self.__getattribute__("_connect_" + method)()
             self.method = method
@@ -477,13 +499,15 @@ class GCSFileSystem(AsyncFileSystem):
         self, method, path, *args, json_out=False, info_out=False, **kwargs
     ):
         logger.debug(f"{method.upper()}: {path}, {args}, {kwargs}")
-        self.maybe_refresh()
-        path, jsonin, datain, headers, kwargs = self._get_args(path, *args, **kwargs)
 
         for retry in range(self.retries):
             try:
                 if retry > 0:
                     await asyncio.sleep(min(random.random() + 2 ** (retry - 1), 32))
+                self.maybe_refresh()
+                path, jsonin, datain, headers, kwargs = self._get_args(
+                    path, *args, **kwargs
+                )
                 async with self.session.request(
                     method=method,
                     url=path,
@@ -516,12 +540,12 @@ class GCSFileSystem(AsyncFileSystem):
                     msg = "Bucket is requester pays. Set `requester_pays=True` when creating the GCSFileSystem."
                     raise ValueError(msg) from e
                 if retry == self.retries - 1:
-                    logger.exception("_call out of retries on exception: %s", e)
+                    logger.exception("_call out of retries on exception: %s" % e)
                     raise e
                 if is_retriable(e):
-                    logger.debug("_call retrying after exception: %s", e)
+                    logger.debug("_call retrying after exception: %s" % e)
                     continue
-                logger.exception("_call non-retriable exception: %s", e)
+                logger.exception("_call non-retriable exception: %s" % e)
                 raise e
         if json_out:
             return json
@@ -535,7 +559,12 @@ class GCSFileSystem(AsyncFileSystem):
     @property
     def buckets(self):
         """Return list of available project buckets."""
-        return [b["name"] for b in sync(self.loop, self._list_buckets())]
+        return [
+            b["name"]
+            for b in sync(
+                self.loop, self._list_buckets(), callback_timeout=self.callback_timeout
+            )
+        ]
 
     @staticmethod
     def _process_object(bucket, object_metadata):
@@ -572,13 +601,26 @@ class GCSFileSystem(AsyncFileSystem):
             # listing.
             raise FileNotFoundError(path)
 
-        result = self._process_object(
-            bucket, await self._call("GET", "b/{}/o/{}", bucket, key, json_out=True)
-        )
+        res = None
+        # Work around various permission settings. Prefer an object get (storage.objects.get), but
+        # fall back to a bucket list + filter to object name (storage.objects.list).
+        try:
+            res = await self._call("GET", "b/{}/o/{}", bucket, key, json_out=True)
+        except OSError as e:
+            if not str(e).startswith("Forbidden"):
+                raise
+            resp = await self._call(
+                "GET", "b/{}/o/", bucket, json_out=True, prefix=key, maxResults=1
+            )
+            for item in resp.get("items", []):
+                if item["name"] == key:
+                    res = item
+                    break
+            if res is None:
+                raise FileNotFoundError(path)
+        return self._process_object(bucket, res)
 
-        return result
-
-    async def _list_objects(self, path):
+    async def _list_objects(self, path, prefix=""):
         bucket, key = self.split_path(path)
         path = path.rstrip("/")
 
@@ -595,7 +637,7 @@ class GCSFileSystem(AsyncFileSystem):
             if key:
                 raise
 
-        items, prefixes = await self._do_list_objects(path)
+        items, prefixes = await self._do_list_objects(path, prefix=prefix)
 
         pseudodirs = [
             {
@@ -613,13 +655,16 @@ class GCSFileSystem(AsyncFileSystem):
             else:
                 return []
         out = items + pseudodirs
-        self.dircache[path] = out
+        # Don't cache prefixed/partial listings
+        if not prefix:
+            self.dircache[path] = out
         return out
 
-    async def _do_list_objects(self, path, max_results=None, delimiter="/"):
+    async def _do_list_objects(self, path, max_results=None, delimiter="/", prefix=""):
         """Object listing for the given {bucket}/{prefix}/ path."""
-        bucket, prefix = self.split_path(path)
-        prefix = None if not prefix else prefix.rstrip("/") + "/"
+        bucket, _path = self.split_path(path)
+        _path = "" if not _path else _path.rstrip("/") + "/"
+        prefix = f"{_path}{prefix}" or None
 
         prefixes = []
         items = []
@@ -781,7 +826,8 @@ class GCSFileSystem(AsyncFileSystem):
             return await self._get_object(path)
         except FileNotFoundError:
             pass
-        out = await self._ls(path, detail=True, **kwargs)
+        kwargs["detail"] = True  # Force to true for info
+        out = await self._ls(path, **kwargs)
         out0 = [o for o in out if o["name"].rstrip("/") == path]
         if out0:
             # exact hit
@@ -798,14 +844,26 @@ class GCSFileSystem(AsyncFileSystem):
         else:
             raise FileNotFoundError(path)
 
-    async def _ls(self, path, detail=False, **kwargs):
+    def glob(self, path, prefix="", **kwargs):
+        if not prefix:
+            # Identify pattern prefixes. Ripped from fsspec.spec.AbstractFileSystem.glob and matches
+            # the glob.has_magic patterns.
+            indstar = path.find("*") if path.find("*") >= 0 else len(path)
+            indques = path.find("?") if path.find("?") >= 0 else len(path)
+            indbrace = path.find("[") if path.find("[") >= 0 else len(path)
+
+            ind = min(indstar, indques, indbrace)
+            prefix = path[:ind].split("/")[-1]
+        return super().glob(path, prefix=prefix, **kwargs)
+
+    async def _ls(self, path, detail=False, prefix="", **kwargs):
         """List objects under the given '/{bucket}/{prefix} path."""
         path = self._strip_protocol(path).rstrip("/")
 
         if path in ["/", ""]:
             out = await self._list_buckets()
         else:
-            out = await self._list_objects(path)
+            out = await self._list_objects(path, prefix=prefix)
 
         if detail:
             return out
@@ -820,10 +878,14 @@ class GCSFileSystem(AsyncFileSystem):
         object = quote_plus(object)
         return u.format(bucket, object)
 
-    async def _cat_file(self, path):
+    async def _cat_file(self, path, start=None, end=None):
         """ Simple one-shot get of file data """
         u2 = self.url(path)
-        headers, out = await self._call("GET", u2)
+        if start or end:
+            head = {"Range": "bytes=%i-%s" % (start or 0, end - 1 if end else "")}
+        else:
+            head = {}
+        headers, out = await self._call("GET", u2, headers=head)
         return out
 
     def getxattr(self, path, attr):
@@ -834,7 +896,7 @@ class GCSFileSystem(AsyncFileSystem):
     async def _setxattrs(
         self, path, content_type=None, content_encoding=None, **kwargs
     ):
-        """ Set/delete/add writable metadata attributes
+        """Set/delete/add writable metadata attributes
 
         Parameters
         ---------
@@ -892,8 +954,7 @@ class GCSFileSystem(AsyncFileSystem):
     merge = sync_wrapper(_merge)
 
     async def _cp_file(self, path1, path2, acl=None, **kwargs):
-        """Duplicate remote file
-        """
+        """Duplicate remote file"""
         b1, k1 = self.split_path(path1)
         b2, k2 = self.split_path(path2)
         out = await self._call(
@@ -1014,12 +1075,10 @@ class GCSFileSystem(AsyncFileSystem):
                 out = await upload_chunk(
                     self, location, bit, offset, size, content_type
                 )
-        if consistency == "size":
-            assert int(out["size"]) == size
-        elif consistency == "md5":
-            md = md5()
-            md.update(data)
-            assert out["md5Hash"] == b64encode(md.digest()).decode()
+
+        checker = get_consistency_checker(consistency)
+        checker.update(data)
+        checker.validate_json_response(out)
         self.invalidate_cache(self._parent(path))
 
     async def _put_file(
@@ -1036,11 +1095,11 @@ class GCSFileSystem(AsyncFileSystem):
         if os.path.isdir(lpath):
             return
         consistency = consistency or self.consistency
+        checker = get_consistency_checker(consistency)
         bucket, key = self.split_path(rpath)
         with open(lpath, "rb") as f0:
             size = f0.seek(0, 2)
             f0.seek(0)
-            md = md5()
             if size < 5 * 2 ** 20:
                 return await simple_upload(
                     self,
@@ -1064,13 +1123,9 @@ class GCSFileSystem(AsyncFileSystem):
                         self, location, bit, offset, size, content_type
                     )
                     offset += len(bit)
-                    if consistency == "md5":
-                        # check this chunk against X-Range-MD5 response header?
-                        md.update(bit)
-            if consistency == "size":
-                assert int(out["size"]) == size
-            elif consistency == "md5":
-                assert out["md5Hash"] == b64encode(md.digest()).decode()
+                    checker.update(bit)
+
+            checker.validate_json_response(out)
             self.invalidate_cache(self._parent(rpath))
 
     async def _isdir(self, path):
@@ -1079,15 +1134,56 @@ class GCSFileSystem(AsyncFileSystem):
         except IOError:
             return False
 
-    def find(self, path, withdirs=False, detail=False, **kwargs):
+    def find(self, path, withdirs=False, detail=False, prefix="", **kwargs):
         path = self._strip_protocol(path)
         bucket, key = self.split_path(path)
-        out, _ = sync(self.loop, self._do_list_objects, path, delimiter=None)
+        out, _ = sync(
+            self.loop,
+            self._do_list_objects,
+            path,
+            delimiter=None,
+            prefix=prefix,
+            callback_timeout=self.callback_timeout,
+        )
         if not out and key:
             try:
-                out = [sync(self.loop, self._get_object, path)]
+                out = [
+                    sync(
+                        self.loop,
+                        self._get_object,
+                        path,
+                        callback_timeout=self.callback_timeout,
+                    )
+                ]
             except FileNotFoundError:
                 out = []
+        dirs = []
+        sdirs = set()
+        cache_entries = {}
+        for o in out:
+            par = self._parent(o["name"])
+            if par not in sdirs:
+                sdirs.add(par)
+                dirs.append(
+                    {
+                        "Key": self.split_path(par)[1],
+                        "Size": 0,
+                        "name": par,
+                        "StorageClass": "DIRECTORY",
+                        "type": "directory",
+                        "size": 0,
+                    }
+                )
+            # Don't cache "folder-like" objects (ex: "Create Folder" in GCS console) to prevent
+            # masking subfiles in subsequent requests.
+            if not o["name"].endswith("/"):
+                cache_entries.setdefault(par, [])
+                cache_entries[par].append(o)
+        self.dircache.update(cache_entries)
+
+        if withdirs:
+            out = sorted(out + dirs, key=lambda x: x["name"])
+
         if detail:
             return {o["name"]: o for o in out}
         return [o["name"] for o in out]
@@ -1111,37 +1207,34 @@ class GCSFileSystem(AsyncFileSystem):
             kwargs["userProject"] = user_project
 
         async with self.session.get(
-            url=u2, params=kwargs, headers=headers, timeout=self.requests_timeout,
+            url=u2,
+            params=kwargs,
+            headers=headers,
+            timeout=self.requests_timeout,
         ) as r:
             r.raise_for_status()
-            if self.consistency == "md5":
-                md = md5()
+            checker = get_consistency_checker(consistency)
 
             os.makedirs(os.path.dirname(lpath), exist_ok=True)
             with open(lpath, "wb") as f2:
-                size = 0
                 while True:
                     data = await r.content.read(4096)
                     if not data:
                         break
                     f2.write(data)
-                    if consistency == "md5":
-                        md.update(data)
-                    size = f2.tell()
+                    checker.update(data)
 
-            if consistency == "size":
-                assert r.content_length == size
-            if consistency == "md5":
-                dig = [
-                    bit.split("=")[1]
-                    for bit in r.headers["X-Goog-Hash"].split(",")
-                    if bit.split("=")[0] == "md5"
-                ][0]
-                assert b64encode(md.digest()).decode().rstrip("=") == dig
+            checker.validate_http_response(r)
 
     def rm(self, path, recursive=False, batchsize=20):
         paths = self.expand_path(path, recursive=recursive)
-        sync(self.loop, self._rm, paths, batchsize=batchsize)
+        sync(
+            self.loop,
+            self._rm,
+            paths,
+            callback_timeout=self.callback_timeout,
+            batchsize=batchsize,
+        )
 
     def _open(
         self,
@@ -1232,14 +1325,9 @@ class GCSFileSystem(AsyncFileSystem):
                 raise RuntimeError(msg)
         else:
             if headers is not None and "X-Goog-Hash" in headers:
-                # if header includes md5 hash, check that data matches
-                bits = headers["X-Goog-Hash"].split(",")
-                for bit in bits:
-                    key, val = bit.split("=", 1)
-                    if key == "md5":
-                        md = b64decode(val)
-                        if md5(content).digest() != md:
-                            raise ChecksumError("Checksum failure")
+                checker = MD5Checker()
+                checker.update(content)
+                checker.validate_headers(headers)
 
 
 GCSFileSystem.load_tokens()
@@ -1259,6 +1347,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         consistency="md5",
         metadata=None,
         content_type=None,
+        callback_timeout=None,
         **kwargs,
     ):
         """
@@ -1277,16 +1366,18 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             ACL to apply, if any, one of ``ACLs``. New files are normally
             "bucketownerfullcontrol", but a default can be configured per
             bucket.
-        consistency: str, 'none', 'size', 'md5'
+        consistency: str, 'none', 'size', 'md5', 'crc32c'
             Check for success in writing, applied at file close.
             'size' ensures that the number of bytes reported by GCS matches
             the number we wrote; 'md5' does a full checksum. Any value other
-            than 'size' or 'md5' is assumed to mean no checking.
+            than 'size' or 'md5' or 'crc32' is assumed to mean no checking.
         content_type: str
             default is `application/octet-stream`. See the list of available
             content types at https://www.iana.org/assignments/media-types/media-types.txt
         metadata: dict
             Custom metadata, in key/value pairs, added at file creation
+        callback_timeout: int
+            Timeout seconds for the asynchronous callback.
         """
         super().__init__(
             gcsfs,
@@ -1306,10 +1397,9 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         self.key = key
         self.metadata = metadata
         self.acl = acl
-        self.consistency = consistency
+        self.checker = get_consistency_checker(consistency)
         self.content_type = content_type or "application/octet-stream"
-        if self.consistency == "md5":
-            self.md5 = md5()
+        self.callback_timeout = callback_timeout
         if mode == "wb":
             if self.blocksize < GCS_MIN_BLOCK_SIZE:
                 warnings.warn("Setting block size to minimum value, 2**18")
@@ -1325,7 +1415,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
         return self.details["mediaLink"]
 
     def _upload_chunk(self, final=False):
-        """ Write one part of a multi-block file upload
+        """Write one part of a multi-block file upload
 
         Parameters
         ----------
@@ -1362,32 +1452,19 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             end = int(headers["Range"].split("-")[1])
             shortfall = (self.offset + l - 1) - end
             if shortfall:
-                if self.consistency == "md5":
-                    self.md5.update(data[:-shortfall])
-                    # check this chunk against X-Range-MD5 response header?
+                self.checker.update(data[:-shortfall])
                 self.buffer = io.BytesIO(data[-shortfall:])
                 self.buffer.seek(shortfall)
                 self.offset += l - shortfall
                 return False
             else:
-                if self.consistency == "md5":
-                    # check this chunk against X-Range-MD5 response header?
-                    self.md5.update(data)
+                self.checker.update(data)
         elif l:
             #
             assert final, "Response looks like upload is over"
             j = json.loads(contents)
-            size, md5 = int(j["size"]), j["md5Hash"]
-            if self.consistency == "size":
-                # update offset with final chunk of data
-                self.offset += l
-                assert size == self.buffer.tell() + self.offset, "Size mismatch"
-            if self.consistency == "md5":
-                # update md5 with final chunk of data
-                self.md5.update(data)
-                assert (
-                    b64encode(self.md5.digest()) == md5.encode()
-                ), "MD5 checksum failed"
+            self.checker.update(data)
+            self.checker.validate_json_response(j)
         else:
             assert final, "Response looks like upload is over"
         return True
@@ -1407,6 +1484,7 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             self.key,
             self.content_type,
             self.metadata,
+            callback_timeout=self.callback_timeout,
         )
 
     def discard(self):
@@ -1439,10 +1517,11 @@ class GCSFile(fsspec.spec.AbstractBufferedFile):
             self.metadata,
             self.consistency,
             self.content_type,
+            callback_timeout=self.callback_timeout,
         )
 
     def _fetch_range(self, start=None, end=None):
-        """ Get data from GCS
+        """Get data from GCS
 
         start, end : None or integers
             if not both None, fetch only given range
@@ -1492,7 +1571,10 @@ async def initiate_upload(
         headers={"X-Upload-Content-Type": content_type},
     )
     loc = headers["Location"]
-    return loc[0] if isinstance(loc, list) else loc  # <- for CVR responses
+    out = loc[0] if isinstance(loc, list) else loc  # <- for CVR responses
+    if len(str(loc)) < 20:
+        logger.error("Location failed: %s" % headers)
+    return out
 
 
 async def simple_upload(
@@ -1504,6 +1586,7 @@ async def simple_upload(
     consistency=None,
     content_type="application/octet-stream",
 ):
+    checker = get_consistency_checker(consistency)
     path = "https://www.googleapis.com/upload/storage/v1/b/%s/o" % quote_plus(bucket)
     metadata = {"name": key}
     if metadatain is not None:
@@ -1528,10 +1611,5 @@ async def simple_upload(
         data=data,
         json_out=True,
     )
-    size, mdback = int(j["size"]), j["md5Hash"]
-    if consistency == "size":
-        assert size == len(datain)
-    if consistency == "md5":
-        md = md5()
-        md.update(datain)
-        assert b64encode(md.digest()) == mdback.encode(), "MD5 checksum failed"
+    checker.update(datain)
+    checker.validate_json_response(j)
