@@ -34,6 +34,7 @@ from gcsfs.tests.utils import (
     tmpfile,
 )
 from gcsfs.core import GCSFileSystem, quote_plus
+import gcsfs.checkers
 from gcsfs import __version__ as version
 
 pytestmark = pytest.mark.usefixtures("token_restore")
@@ -279,11 +280,18 @@ def test_gcs_glob():
         assert fn in gcs.glob(TEST_BUCKET + "/nested/file*")
         assert fn in gcs.glob(TEST_BUCKET + "/*/*")
         assert fn in gcs.glob(TEST_BUCKET + "/**")
+        assert fn in gcs.glob(TEST_BUCKET + "/**1")
         assert all(
             f in gcs.find(TEST_BUCKET)
             for f in gcs.glob(TEST_BUCKET + "/nested/*")
             if gcs.isfile(f)
         )
+        # Ensure the glob only fetches prefixed folders
+        gcs.dircache.clear()
+        gcs.glob(TEST_BUCKET + "/nested**1")
+        assert all(d.startswith(TEST_BUCKET + "/nested") for d in gcs.dircache)
+        gcs.glob(TEST_BUCKET + "/test*")
+        assert TEST_BUCKET + "/test" in gcs.dircache
 
 
 @my_vcr.use_cassette(match=["all"])
@@ -355,6 +363,49 @@ def test_copy():
         assert gcs.cat(fn) == gcs.cat(fn + "2")
 
 
+@my_vcr.use_cassette(match=["all"])
+def test_copy_recursive():
+    with gcs_maker(True) as gcs:
+        src = TEST_BUCKET + "/nested"
+        dest = TEST_BUCKET + "/dest"
+        gcs.copy(src, dest, recursive=True)
+        for fn in gcs.ls(dest):
+            if not gcs.isdir(fn):
+                assert gcs.cat(fn) == gcs.cat(fn.replace("dest", "nested"))
+
+
+@my_vcr.use_cassette(match=["all"])
+def test_copy_errors():
+    with gcs_maker(True) as gcs:
+
+        src = TEST_BUCKET + "/test/"
+        file1 = TEST_BUCKET + "/test/accounts.1.json"
+        file2 = TEST_BUCKET + "/test/accounts.2.json"
+        dne = TEST_BUCKET + "/test/notafile"
+        dest1 = TEST_BUCKET + "/dest1/"
+        dest2 = TEST_BUCKET + "/dest2/"
+
+        # Non recursive should raise an error unless we specify ignore
+        with pytest.raises(FileNotFoundError):
+            gcs.copy([file1, file2, dne], dest1)
+
+        gcs.copy([file1, file2, dne], dest1, on_error="ignore")
+        assert gcs.ls(dest1) == [dest1 + "accounts.1.json", dest1 + "accounts.2.json"]
+
+        # Recursive should raise an error only if we specify raise
+        # the patch simulates the filesystem finding a file that does not exist in the directory
+        current_files = gcs.expand_path(src, recursive=True)
+        with mock.patch.object(gcs, "expand_path", return_value=current_files + [dne]):
+            with pytest.raises(FileNotFoundError):
+                gcs.copy(src, dest2, recursive=True, on_error="raise")
+
+            gcs.copy(src, dest2, recursive=True)
+            assert gcs.ls(dest2) == [
+                dest2 + "accounts.1.json",
+                dest2 + "accounts.2.json",
+            ]
+
+
 @my_vcr.use_cassette(match=["all"])
 def test_move():
     with gcs_maker(True) as gcs:
@@ -365,10 +416,22 @@ def test_move():
         assert not gcs.exists(fn)
 
 
+@my_vcr.use_cassette(match=["all"])
+def test_cat_file():
+    with gcs_maker(True) as gcs:
+        fn = TEST_BUCKET + "/test/accounts.1.json"
+        data = gcs.cat_file(fn)
+        assert data[1:10] == gcs.cat_file(fn, start=1, end=10)
+        assert data[1:] == gcs.cat_file(fn, start=1)
+        assert data[:1] == gcs.cat_file(fn, end=1)
+
+
 @pytest.mark.skipif(ON_VCR, reason="async fail")
 @my_vcr.use_cassette(match=["all"])
-@pytest.mark.parametrize("consistency", [None, "size", "md5"])
+@pytest.mark.parametrize("consistency", [None, "size", "md5", "crc32c"])
 def test_get_put(consistency):
+    if consistency == "crc32c" and gcsfs.checkers.crcmod is None:
+        pytest.skip("No CRC")
     with gcs_maker(True) as gcs:
         gcs.consistency = consistency
         with tmpfile() as fn:
@@ -414,6 +477,33 @@ def test_get_put_recursive(protocol):
             )
 
 
+@pytest.mark.skipif(ON_VCR, reason="async fail")
+@pytest.mark.parametrize("protocol", ["", "gs://", "gcs://"])
+@my_vcr.use_cassette(match=["all"])
+def test_get_put_file_in_dir(protocol):
+    with gcs_maker(True) as gcs:
+        with tempdir() as dn:
+            gcs.get(protocol + TEST_BUCKET + "/test/", dn + "/temp_dir", recursive=True)
+            # there is now in local directory:
+            # dn+'/temp_dir/accounts.1.json'
+            # dn+'/temp_dir/accounts.2.json'
+            data1 = files["test/accounts.1.json"]
+            assert open(dn + "/temp_dir/accounts.1.json", "rb").read() == data1
+            gcs.put(
+                dn + "/temp_dir/accounts.1.json",
+                protocol + TEST_BUCKET + "/temp_dir/",
+                recursive=True,
+            )
+            # there is now in remote directory:
+            # protocol+TEST_BUCKET+'/temp_dir/accounts.1.json'
+            assert gcs.du(protocol + TEST_BUCKET + "/temp_dir/accounts.1.json") == len(
+                data1
+            )
+            assert (
+                gcs.cat(protocol + TEST_BUCKET + "/temp_dir/accounts.1.json") == data1
+            )
+
+
 @my_vcr.use_cassette(match=["all"])
 def test_errors():
     with gcs_maker() as gcs:
@@ -869,14 +959,14 @@ def test_validate_response():
     with pytest.raises(HttpError) as e:
         gcs.validate_response(503, b"", None, "/path")
     assert e.value.code == 503
-    assert e.value.message == ""
+    assert e.value.message == ", 503"
 
     # HttpError with JSON body
     j = {"error": {"code": 503, "message": b"Service Unavailable"}}
     with pytest.raises(HttpError) as e:
         gcs.validate_response(503, None, j, "/path")
     assert e.value.code == 503
-    assert e.value.message == b"Service Unavailable"
+    assert e.value.message == b"Service Unavailable, 503"
 
     # 403
     j = {"error": {"message": "Not ok"}}
@@ -895,3 +985,104 @@ def test_validate_response():
     md5 = repr(base64.b64encode(hashlib.md5(b"foo").digest()))[2:-1]
     with pytest.raises(ChecksumError):
         gcs.validate_response(0, b"f", None, "/path", {"X-Goog-Hash": f"md5={md5}"})
+
+
+@my_vcr.use_cassette(match=["all"])
+@pytest.mark.parametrize(
+    ["file_path", "validate_get_error", "validate_list_error", "expected_error"],
+    [
+        (
+            "/missing",
+            FileNotFoundError,
+            None,  # Not called
+            FileNotFoundError,
+        ),
+        (
+            "/missing",
+            OSError("Forbidden"),
+            FileNotFoundError,
+            FileNotFoundError,
+        ),
+        (
+            "/2014-01-01.csv",
+            None,
+            None,
+            None,
+        ),
+        (
+            "/2014-01-01.csv",
+            OSError("Forbidden"),
+            None,
+            None,
+        ),
+    ],
+    ids=[
+        "missing_with_get_perms",
+        "missing_with_list_perms",
+        "existing_with_get_perms",
+        "existing_with_list_perms",
+    ],
+)
+def test_metadata_read_permissions(
+    file_path, validate_get_error, validate_list_error, expected_error
+):
+    with gcs_maker(True) as gcs:
+        _validate_response = gcs.validate_response
+
+        def validate_response(self, status, content, json, path, headers=None):
+            if path.endswith(f"/o{file_path}") and validate_get_error is not None:
+                raise validate_get_error
+            if path.endswith("/o/") and validate_list_error is not None:
+                raise validate_list_error
+            _validate_response(status, content, json, path, headers=None)
+
+        gcs.validate_response = validate_response.__get__(gcs)
+
+        if expected_error is None:
+            gcs.ls(TEST_BUCKET + file_path)
+            gcs.info(TEST_BUCKET + file_path)
+            assert gcs.exists(TEST_BUCKET + file_path)
+        else:
+            with pytest.raises(expected_error):
+                gcs.ls(TEST_BUCKET + file_path)
+            with pytest.raises(expected_error):
+                gcs.info(TEST_BUCKET + file_path)
+            assert gcs.exists(TEST_BUCKET + file_path) is False
+
+
+@my_vcr.use_cassette(match=["all"])
+def test_ls_prefix_cache():
+    with gcs_maker(True) as gcs:
+        gcs.touch(f"gs://{TEST_BUCKET}/a/file1")
+        gcs.touch(f"gs://{TEST_BUCKET}/a/file2")
+
+        gcs.ls(f"gs://{TEST_BUCKET}/", prefix="a/file")
+        gcs.info(f"gs://{TEST_BUCKET}/a/file1")
+
+
+@my_vcr.use_cassette(match=["all"])
+def test_placeholder_dir_cache_validity():
+    with gcs_maker(True) as gcs:
+        gcs.touch(f"gs://{TEST_BUCKET}/a/")
+        gcs.touch(f"gs://{TEST_BUCKET}/a/file")
+        gcs.touch(f"gs://{TEST_BUCKET}/b")
+
+        gcs.find(f"gs://{TEST_BUCKET}/a/")
+        gcs.info(f"gs://{TEST_BUCKET}/b")
+
+
+@my_vcr.use_cassette(match=["all"])
+def test_zero_cache_timeout():
+    with gcs_maker(True, cache_timeout=0) as gcs:
+        gcs.touch(f"gs://{TEST_BUCKET}/a/file")
+        gcs.find(f"gs://{TEST_BUCKET}/a/")
+        gcs.info(f"gs://{TEST_BUCKET}/a/file")
+        gcs.ls(f"gs://{TEST_BUCKET}/a/")
+
+        # The _times entry and exception below should only be present after
+        # https://github.com/intake/filesystem_spec/pull/513.
+        if f"{TEST_BUCKET}/a" not in gcs.dircache._times:
+            pytest.skip("fsspec version too early")
+
+        with pytest.raises(KeyError):
+            gcs.dircache[f"{TEST_BUCKET}/a"]
